#include <CGAL/basic.h>
#include <CGAL/Voronoi_cell_of_line_3.h>
#include <boost/variant.hpp>

// CURRENTLY THIS IS JUST RENAME of the VDOL class. 
// IT SHOULD TAKE A VDOL AND PROVIDE THE CURRENT INTERFACE 
// NEEDED TO GENERATE A MESH 



#ifndef CGAL_VDOL_TO_LABELED_FUNCTION_WRAPPER_H
#define CGAL_VDOL_TO_LABELED_FUNCTION_WRAPPER_H

namespace CGAL {


template <class SVCET_3, class K>
class VDOL_to_labeled_function_wrapper{
private: 
  typedef VDOL_to_labeled_function_wrapper<SVCET_3,K> Self; 
  typedef Self VDOL3;
public:
  typedef typename K::Point_3   Point_3;
  typedef typename K::Line_3    Line_3;
  typedef typename K::Ray_3     Ray_3;
  typedef typename K::Segment_3 Segment_3; 
  typedef typename K::Sphere_3 Sphere_3;

private:
  typedef typename K::Vector_3 Vector_3;

  typedef typename SVCET_3::Linear_kernel EK;
  typedef typename EK::Point_3  EPoint_3;
  typedef typename EK::Vector_3 EVector_3;
  typedef typename EK::Line_3   ELine_3;
  typedef typename EK::FT Rational;
  
  
  typedef typename SVCET_3::Curved_kernel_2 CK_2;

  typedef typename SVCET_3::Point_2 Point_2;

  typedef typename SVCET_3::Poly_int_3 Poly_int_3; 
  typedef typename SVCET_3::Poly_int_2 Poly_int_2; 
  typedef typename SVCET_3::Poly_int_1 Poly_int_1; 

  typedef typename SVCET_3::Poly_rat_3 Poly_rat_3; 
  typedef typename SVCET_3::Poly_rat_2 Poly_rat_2; 
  typedef typename SVCET_3::Poly_rat_1 Poly_rat_1; 
  
  typedef typename SVCET_3::Xy_monotone_surface_3 Xy_monotone_surface_3;

  typedef typename CK_2::Curve_kernel_2 AK_2;
  typedef typename AK_2::Coordinate_1 Coordinate_1 ;
  typedef typename AK_2::Solve_1 Solve_1; 
  typedef typename AK_2::Coefficient Integer; 
  
  typedef typename CGAL::Envelope_diagram_2<SVCET_3> Envelope_diagram_2;
  typedef  Arr_naive_point_location<Envelope_diagram_2> Point_location;
  typedef Voronoi_cell_of_line_3<SVCET_3,K> VCOL_3;
  
    // BEGIN TYPEDEFS FOR CONCEPT MeshDomain_3
public:
  typedef K R; 
  typedef int Subdomain_index;
  typedef std::pair<int,int> Surface_index; 
  typedef boost::variant<Subdomain_index,Surface_index> Index; 
  typedef boost::optional<Subdomain_index> Subdomain; 
  typedef boost::optional<Surface_index> Surface_patch; ;
  typedef CGAL::cpp0x::tuple<Point_3,Index,int> Intersection ;

  Index index_from_surface_index(const Surface_index& i) const 
  {return Index(i);}
  Index index_from_subdomain_index(const Subdomain_index& i) const 
  {return Index(i);}
  Surface_index surface_index(const Index& i) const 
  {return boost::get<Surface_index>(i);}
  Subdomain_index subdomain_index(const Index& i) const 
  {return boost::get<Subdomain_index>(i);}
  // END TYPEDEFS FOR CONCEPT MeshDomain_3 

private: 
// ##########################################

  std::vector<VCOL_3>  m_cells; 
  std::vector<ELine_3> m_lines;
  std::vector<std::pair<Point_3,double> > m_wpoints;
 
  

public:
  const std::vector<std::pair<Point_3,double> >& wpoints() const {return m_wpoints;}
  std::vector<std::pair<Point_3,double> >& wpoints(){return m_wpoints;}
  
  template <class InputIterator>
  VDOL_to_labeled_function_wrapper(InputIterator begin, InputIterator end)
    :m_lines(begin,end)
  {
    std::cout << " number of lines " <<  m_lines.size() << std::endl;

    for(int i = 0; i < m_lines.size(); i++){
      //for(int i = 0; i < 1 ; i++){
      std::cout << " cell number " <<  i  << std::endl;
      std::swap(m_lines[0],m_lines[i]);

      int seed = 5; 
      bool done = false; 
      while(!done){
        std::cout << "try with seed " << seed << std::flush; 
        try{
          m_cells.push_back(VCOL_3(seed,m_lines.begin(),m_lines.end()));
          done = true ;
        }catch(CGAL::VDOL_3::Event_at_discontinuity_exception& e){
          std::cout << "catch:" << std::endl 
                    << e.what() << std::endl;
          seed++; 
          if(seed > 10) {
            std::cout << "infinite try/catch unexpected special case ? " 
                      << std::endl;
            assert(false);
          }
        } 
      }
    }
    init_points();
  }
  
  void init_points(){
    std::cout << "INIT POINTS" << std::endl; 

    std::vector<std::pair<Point_3,double > > wpoints_raw;
    for(int i = 0; i < m_cells.size();i++){
      std::copy(m_cells[i].wpoints().begin(),m_cells[i].wpoints().end(),std::back_inserter(wpoints_raw));
    }
    
    K kernel; 
    typename K::Compute_squared_distance_3 sdist = kernel.compute_squared_distance_3_object();    

    for(int i = 0; i < wpoints_raw.size(); i++){
      if( slength(wpoints_raw[i].first) < VOL_CLIPPING_SPHERE + 0.1){
        bool insert = true; 
        for(int j = 0; j < m_wpoints.size() && insert ; j++){
          if(CGAL::to_double(sdist(wpoints_raw[i].first,m_wpoints[j].first)) < VOL_POINT_DIST){
            insert = false ;
          }          
        }
        if(insert) m_wpoints.push_back(wpoints_raw[i]);
      }
    }
    
  }
  

  double slength(const Point_3& p) const {return p.x()*p.x()+ p.y()*p.y()+p.z()*p.z(); }
  Sphere_3 bounding_sphere() const {return Sphere_3(Point_3(0,0,0),VOL_BOUNDING_SPHERE);}
  Sphere_3 clipping_sphere() const {return Sphere_3(Point_3(0,0,0),VOL_CLIPPING_SPHERE);}

  
  const VCOL_3& containing_cell (const Point_3& p) const {
    for(int i = 0; i < m_cells.size(); i++){
      if (m_cells[i](p)==1) 
        return m_cells[i];
    }
  }
  int label(const Point_3& p) const {
    for(int i = 0; i < m_cells.size(); i++){
      if (m_cells[i](p)==1) 
        return i+1 + m_cells[i].compute_segment_index(p) ;
    }
    return 0;
  }
  
  boost::optional<Intersection> intersect (const Line_3 line) const {
    std::cout << " intersect with Line" << std::endl;  
    CGAL::Bbox_3 bbox = typename K::Construct_bbox_3()(bounding_sphere());
    const Object clipped = CGAL::intersection(line, bbox);
    if ( const Segment_3* s = object_cast<Segment_3>(&clipped) )
      return intersect(s->source(), s->target());
    return  boost::optional<Intersection>(); 
    
  }  

  boost::optional<Intersection> intersect (const Ray_3& ray) const {
    std::cout << " intersect with Ray " << std::endl;  
    std::cout << ray.source() << std::endl; 
    CGAL::Bbox_3 bbox = typename K::Construct_bbox_3()(bounding_sphere()); 
    std::cout << bbox << std::endl;
    const Object clipped = CGAL::intersection(ray, bbox);
    if ( const Segment_3* s = object_cast<Segment_3>(&clipped) )
      return intersect(s->source(), s->target());
    std::cout << (clipped.type().name()) << std::endl; 
    assert(false); 
    return  boost::optional<Intersection>(); 
    
  }

  boost::optional<Intersection> intersect (const Segment_3& s) const{ 
    std::cout << " intersect with Segment " << std::endl;  
    std::cout << " source: " << s.source() << std::endl;  
    std::cout << " target: " << s.target() << std::endl;  
    return intersect(s.source(),s.target());
  }
  
  boost::optional<Intersection> intersect (const Point_3& source, const Point_3& target) const {
    
    // todo: move this somewhere more general 
    static const Poly_rat_3 x = CGAL::shift(Poly_rat_3(1),1,0); 
    static const Poly_rat_3 y = CGAL::shift(Poly_rat_3(1),1,1); 
    static const Poly_rat_3 z = CGAL::shift(Poly_rat_3(1),1,2);
    static const Poly_rat_1 t = CGAL::shift(Poly_rat_1(1),1,0); 


    CGAL::set_pretty_mode(std::cout); 
    std::cout << " ---------------- "  << std::endl;


    // std::cout << source << std::endl;
    // std::cout << target << std::endl;

    //   int scell_index = this->vdol_3()->label(source); 
    //int tcell_index = this->vdol_3()->label(target);
    //const VCOL_3& scell = this->vdol_3()->containing_cell(source);
    //const VCOL_3& tcell = this->vdol_3()->containing_cell(target); 
    
    Poly_rat_3 bisector; 

//     if (scell == 0 || tcell == 0 ){
//       bisector = x*x+y*y+z*z-VOL_CLIPPING_SPHERE; 
//     }else{
//       bisector =  VDOL_3::construct_bisector_3<EK>(scell->line_3(),tcell->line_3());
//     }
    
    bisector = x*x+y*y+z*z-VOL_CLIPPING_SPHERE; 
   

    // std::cout << bisector << std::endl; 

    std::vector<Poly_rat_1> l_param; 
    l_param.push_back(Rational(source.x())*(1-t)+Rational(target.x())*t);
    l_param.push_back(Rational(source.y())*(1-t)+Rational(target.y())*t); 
    l_param.push_back(Rational(source.z())*(1-t)+Rational(target.z())*t);

    //std::cout << l_param[0] << std::endl; 
    //std::cout << l_param[1] << std::endl; 
    //std::cout << l_param[2] << std::endl; 
    
    Poly_rat_1 res = CGAL::substitute(bisector,l_param.begin(),l_param.end());
    
    // std::cout << res << std::endl;
    Integer dummy;
    Poly_int_1 res_int; 
    typename CGAL::Fraction_traits<Poly_rat_1>::Decompose()(res,res_int,dummy);
    // std::cout << res_int << std::endl;

    std::vector<typename AK_2::Algebraic_real_1> roots; 
    // Note that the segemnt may touch the sphere. 
    // this is still the old interface as implemented in the MPI AK. 
    AK_2::get_static_instance().solve_1_object()(res_int, std::back_inserter(roots), false );

    double param;
    bool intersects = false;
    typename AK_2::Compare_1 comp =  AK_2::get_static_instance().compare_1_object();
    if(roots.size()>=1){
      //std::cout << to_double(roots[0])<< std::endl;
      if(comp(roots[0],0) != CGAL::SMALLER && comp(roots[0],1) != CGAL::LARGER){
        param = CGAL::to_double(roots[0]); 
        intersects = true ; 
      }
    }
    if(roots.size()==2){
      //std::cout << to_double(roots[1])<< std::endl;
      if(comp(roots[1],0) != CGAL::SMALLER && comp(roots[1],1) != CGAL::LARGER){
        param = CGAL::to_double(roots[1]);  
        intersects = true ; 
      }
    }
    if(intersects){
      Point_3 p (
          source.x()*(1-param)+target.x()*param,
          source.y()*(1-param)+target.y()*param,
          source.z()*(1-param)+target.z()*param);
      
      std::cout << p << std::endl;
      Index index = Index(Surface_index(0,1));
      return  boost::optional<Intersection> (Intersection(p,index,2));
      
    }else{
      return  boost::optional<Intersection>();
    }
  }

 



  // Concept Function label 
  typedef int return_type;
  return_type operator()(typename K::Point_3 p, bool b = true ) const {
    return label(p); 
  }

  // CONCEPT MeshDomain 

private:
  class Functor_base {  
  protected:
    const VDOL3* m_vdol_3;
    const VDOL3* vdol_3() const { return m_vdol_3; }
  public:
    Functor_base(const VDOL3* ptr) : m_vdol_3(ptr) {
      CGAL_precondition(ptr != NULL);
    }  
  };

public:
  class Construct_initial_points: public Functor_base {
    typedef Functor_base Base; 
  public:
    Construct_initial_points(const VDOL3* ptr): Base(ptr) {};

    template <class OutputIterator>
    OutputIterator operator()(OutputIterator& oit) const {
      std::cout << "Construct_initial_points" << std::endl; 
      assert(false);
      return oit; 
    }
    template <class OutputIterator>
    OutputIterator operator()(int n, OutputIterator& oit) const {
      std::cout << "Construct_initial_points"<< n << std::endl; 
      assert(false);
      return oit; 
    }
  };
  
  class Is_in_domain : public Functor_base {
    typedef Functor_base Base; 
  public:
    Is_in_domain(const VDOL3* ptr): Base(ptr) {};
    Subdomain operator()(const Point_3& p) const {
      std::cout << "Is_in_domain" << p << std::endl; 
      if(this->vdol_3()->operator()(p) != 0) return Subdomain(1);
      return Subdomain();
    }
  };

  class Do_intersect_surface: public Functor_base{
    typedef Functor_base Base; 
  public:
    Do_intersect_surface(const VDOL3* ptr): Base(ptr) {};
     
    template <class SRL> // Segment/Ray/Line 
    Surface_patch operator()(const SRL& s) const {
      std::cout << "Do_intersect_surface" << std::endl; 
      boost::optional<Intersection> intersects = this->vdol_3()->intersect(s);
      if(intersects){
        return Surface_patch(boost::get<Surface_index>(CGAL::cpp0x::get<1>(*intersects)));
      }
      return Surface_patch();
    };
  };

  class Construct_intersection: public Functor_base{
    typedef Functor_base Base; 
  public:
    Construct_intersection(const VDOL3* ptr): Base(ptr) {};
    
    template <class SRL> // Segment/Ray/Line 
    Intersection operator()(const SRL& s) const {
      std::cout << "Construct_intersection" << std::endl; 
      boost::optional<Intersection> intersects(this->vdol_3()->intersect(s)); 
      CGAL_precondition(intersects);
      return *intersects; 
    };
  };

#define SNAP_FUNCTOR(_Name,_get_object)                 \
  _Name _get_object() const  { return _Name(this); }    
  
  SNAP_FUNCTOR(Construct_initial_points, Construct_initial_points_object);
  SNAP_FUNCTOR(Is_in_domain,is_in_domain_object);
  SNAP_FUNCTOR(Do_intersect_surface, do_intersect_surface_object);  
  SNAP_FUNCTOR(Construct_intersection, construct_intersection_object);
  
#undef SNAP_FUNCTOR
    
};




} // namespace CGAL 
#endif // CGAL_VDOL_TO_LABELED_FUNCTION_WRAPPER_H 
