// Copyright (c) 2005-2007 Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of CGAL (www.cgal.org); you may redistribute it under
// the terms of the Q Public License version 1.0.
// See the file LICENSE.QPL distributed with CGAL.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// $URL: $
// $Id:
// 
//
// Author(s): Ophir Setter          <ophirset@post.tau.ac.il>
//            Michael Hemmer        <hemmer@mpi-inf.mpg.de>
//
//

#ifndef CGAL_VDL3_CACHE_CONTAINER_H
#define CGAL_VDL3_CACHE_CONTAINER_H

#include <CGAL/VDOL_3/CGAL_SNAP_SVCET_3_TYPEDEFS.h>
#include <CGAL/tuple.h>

CGAL_BEGIN_NAMESPACE
namespace VDOL_3 {

template <typename SVCET_3>
struct Cache_container{  
  CGAL_SNAP_SVCET_3_TYPEDEFS;

  /*! Cache for trisectors begin */
  typedef CGAL::cpp0x::tuple<Sign,Poly_int_3,Poly_int_3> Trisector_cache_key;
  typedef std::list<CGAL::Object>                        Trisector_cache_data;
  typedef std::map<Trisector_cache_key,Trisector_cache_data>  Trisector_cache;
  mutable Trisector_cache m_trisector_cache;
  /*! Cache for trisectors end */

  // KEY: SVCET_3::State, ABOVE/BELOW, arc id, two lines 
  typedef CGAL::cpp0x::tuple<Sign,Sign,int,Poly_int_3,Poly_int_3> Compare_at_arc_cache_key;
  typedef CGAL::Sign Compare_at_arc_cache_data;
  typedef std::map<Compare_at_arc_cache_key,Compare_at_arc_cache_data>  
  Compare_at_arc_cache;
  mutable Compare_at_arc_cache m_compare_at_arc_cache; 

  // projection cache
  typedef std::pair<Poly_int_3, Poly_int_3> Projection_cache_key; 
  typedef boost::optional<Poly_int_2> Projection_cache_data; 
  typedef std::map<Projection_cache_key, Projection_cache_data>  
  Projection_cache;
  mutable Projection_cache m_projection_cache;
};

} // namespace VDOL_3
CGAL_END_NAMESPACE

#endif // CGAL_VDL3_CACHE_CONTAINER_H
