#ifndef CGAL_VORONOI_CELL_OF_LINE_3_H
#define CGAL_VORONOI_CELL_OF_LINE_3_H 

#include <CGAL/basic.h>
#include <CGAL/Timer.h>
#include <CGAL/envelope_3.h>
#include <CGAL/print_envelop_diagram.h>
#include <CGAL/construct_point_2.h>

#include <CGAL/VDOL_3/vol_tools.h>

#include <CGAL/Curved_kernel_via_analysis_2/Curve_renderer_facade.h> 
#include <CGAL/Arr_trapezoid_ric_point_location.h>
#include <CGAL/Arr_landmarks_point_location.h>
#include <CGAL/Arr_walk_along_line_point_location.h>

#include <boost/shared_ptr.hpp>
#include <boost/make_shared.hpp>

#ifndef NDEBUG
#define VOL_POINT_GENERATION   0.5
#define VOL_POINT_DIST   0.61
#define VOL_POINT_WEIGHT 0.6
#define VOL_BOUNDING_SPHERE 110
#define VOL_CLIPPING_SPHERE 100
#else
#if 1
#define VOL_POINT_GENERATION   0.03
#define VOL_POINT_DIST   0.1
#define VOL_POINT_WEIGHT 0
#define VOL_BOUNDING_SPHERE 110 
#define VOL_CLIPPING_SPHERE 100
#else
#define VOL_POINT_GENERATION   0.01
#define VOL_POINT_DIST   0.1
#define VOL_POINT_WEIGHT 0.11
#define VOL_BOUNDING_SPHERE 110 
#define VOL_CLIPPING_SPHERE 100
#endif 
#endif

namespace CGAL {


template <class SVCET_3, class K>
class Voronoi_cell_of_line_3{


public:// typedefs
  typedef K Kernel; 
  typedef int return_type; 
private:
  typedef typename K::Point_3  Point_3;
  typedef typename K::Sphere_3 Sphere_3;
  typedef typename K::Vector_3 Vector_3;

  typedef typename SVCET_3::Linear_kernel EK;
  typedef typename EK::Point_3  EPoint_3;
  typedef typename EK::Vector_3 EVector_3;
  typedef typename EK::Line_3   ELine_3;
  typedef typename EK::FT Rational;
  
  
  typedef typename SVCET_3::Curved_kernel_2 CK_2;

  typedef typename SVCET_3::Point_2 Point_2;
  typedef typename SVCET_3::Line_3  Line_3;

  typedef typename SVCET_3::Poly_int_3 Poly_int_3; 
  typedef typename SVCET_3::Poly_int_2 Poly_int_2; 
  typedef typename SVCET_3::Poly_int_1 Poly_int_1; 

  typedef typename SVCET_3::Poly_rat_3 Poly_rat_3; 
  typedef typename SVCET_3::Poly_rat_2 Poly_rat_2; 
  typedef typename SVCET_3::Poly_rat_1 Poly_rat_1; 
  
  typedef typename SVCET_3::Xy_monotone_surface_3 Xy_monotone_surface_3;

  typedef typename CK_2::Curve_kernel_2 AK_2;
  typedef typename AK_2::Coordinate_1 Coordinate_1 ;

public:
  typedef typename CGAL::Envelope_diagram_2<SVCET_3> Envelope_diagram_2;
private:
  typedef typename Envelope_diagram_2::Topology_traits Topology_traits;
  typedef typename Topology_traits::Vertex Vertex; 
  typedef typename Topology_traits::Halfedge Halfedge;
  typedef typename Topology_traits::Face Face;



  typedef  Arr_naive_point_location<Envelope_diagram_2> Point_location;
  // typedef  Arr_trapezoid_ric_point_location<Envelope_diagram_2> Point_location;
  // typedef  Arr_walk_along_line_point_location<Envelope_diagram_2> Point_location;
  // typedef  Arr_landmarks_point_location<Envelope_diagram_2> Point_location;

class Half_voronoi_cell_of_line_3{
  
private:// member 
  K  m_kernel;
  EK m_ekernel; 
  CK_2 m_ckernel; 
  
  std::vector<Point_3>  m_points; 
  std::vector<Coordinate_1>    m_intersection_values; 
  
  std::vector<Coordinate_1>    m_top_u_values;
  std::vector<Face*>           m_top_faces;
  std::vector<Halfedge*>       m_top_edges;
  
  SVCET_3 m_svcet_3; 
  mutable boost::shared_ptr<Envelope_diagram_2> m_diag; 
  mutable boost::shared_ptr<Point_location> m_pl_diag;


public: // meber access 
  const K& kernel() const {return m_kernel;}
  K& kernel() {return m_kernel;}
  const EK& ekernel() const {return m_ekernel;}
  EK& ekernel() {return m_ekernel;}
  const CK_2& ckernel() const {return m_ckernel;}
  CK_2& ckernel() {return m_ckernel;}

  const std::vector<Point_3>& points() const {return m_points;}
  std::vector<Point_3>& points(){return m_points;}

  const SVCET_3& svcet_3() const { return m_svcet_3;}
  SVCET_3& svcet_3(){return m_svcet_3;}
  
  const Envelope_diagram_2& diag() const { return *m_diag;}
  Envelope_diagram_2& diag(){return *m_diag;}

  long number_of_vertical_asymptotes() const {return m_top_edges.size();}
public:
  
  template <class ForwardIterator>
  Half_voronoi_cell_of_line_3(const SVCET_3& svcet_3_, ForwardIterator begin, ForwardIterator end)
    : m_svcet_3(svcet_3_), m_diag(new Envelope_diagram_2(&svcet_3()))
  {
    init_diagram(begin,end);
    init_point_location();
#ifdef CGAL_VDOL_CELL_SAMPLE_CURVES
    init_points();
#endif
  } 
  
  // this function return true if the point is in a cell associated to this line
  // more over it outputs all lines that are in the less equal distance via oit.
  template<class OutputIterator>
  bool contains(const EPoint_3& p, OutputIterator oit) const {  

    EPoint_3 lp = in_local_coordinates(p);
    CGAL_precondition(lp.z()>=0);
   
    std::vector<Xy_monotone_surface_3> surfaces;
    
    // get all candidates: 
    if(lp.z() != 0){
      CGAL_precondition(lp.z()>0);
      typename SVCET_3::Point_2 pp = project_point(p);
      compute_surfaces(*m_pl_diag,pp,std::back_inserter(surfaces));
    }else if(lp.y() != 0){
      CGAL_precondition(lp.z()==0);
      CGAL_precondition(lp.y() >0);
      int i = distance( m_top_u_values.begin(), lower_bound(m_top_u_values.begin(),m_top_u_values.end(),lp.x()));
      if(i<m_top_u_values.size() && m_top_u_values[i] == lp.x()){ // take the edge at v[i]
        std::copy(m_top_edges[i]->surfaces_begin(),m_top_edges[i]->surfaces_end(),std::back_inserter(surfaces));
      }else{ // take the face to the left of v[i] 
        std::copy(m_top_faces[i]->surfaces_begin(),m_top_faces[i]->surfaces_end(),std::back_inserter(surfaces));
      }
    }else{
      CGAL_precondition(lp.y()==0);
      CGAL_precondition(lp.z()==0);
      // intersections are vertical lines and can be seen at top too. 
      int i = distance( m_top_u_values.begin(),lower_bound(m_top_u_values.begin(),m_top_u_values.end(),lp.x()));
      // if the point is on such a line report all surfaces on the edge
      if(i<m_top_u_values.size() && m_top_u_values[i] == lp.x()){
        std::copy(m_top_edges[i]->surfaces_begin(),m_top_edges[i]->surfaces_end(),std::back_inserter(surfaces));
      }
      // otherwise the point is just on the line and we do nothing 
    }
    
    // filter surfaces:
    Rational min_sdist = CGAL::squared_distance(svcet_3().base_line_3(),p);
    for(int i=0; i<surfaces.size(); i++){
      min_sdist = (std::min)(min_sdist,CGAL::squared_distance(surfaces[i].line(),p));
    }
    
    for(int i=0; i<surfaces.size(); i++){
      if(CGAL::squared_distance(surfaces[i].line(),p) == min_sdist)
        *oit++ = surfaces[i].line();
    }
    
    if (CGAL::squared_distance(svcet_3().base_line_3(),p) == min_sdist){
      *oit++ = svcet_3().base_line_3();
      return true;
    }
    
    return false; 
  }
 
  int compute_segment_index(const Point_3& p) const {
    if(m_intersection_values.size()==0) return 0; 
    
//    EPoint_3 ep = in_local_coordinates(p); 
//    std::cerr << svcet_3().base_line() << std::endl; 
//    std::cerr << p << std::endl;
//    std::cerr << ep << std::endl;

    Rational  t =  in_local_coordinates(p).x() ;

    return distance(
        m_intersection_values.begin(),
        lower_bound(m_intersection_values.begin(),m_intersection_values.end(),t));
  }
  
  void init_point_location(){
    // point location for normal points
    m_pl_diag = boost::shared_ptr<Point_location>(new Point_location(*m_diag));

    // point location for points in H^*
    Vertex   *tv = m_diag->topology_traits()->top_left_vertex();
    Halfedge *ht = tv->halfedge();
    if(VDOL_3::is_vertical(ht)) ht = ht->next()->opposite();
    CGAL_assertion(!VDOL_3::is_vertical(ht));
    CGAL_assertion(ht->vertex() == tv);
    
    ht = ht->opposite()->next()->opposite();
    while(ht->vertex() != m_diag->topology_traits()->top_right_vertex()){
      CGAL_assertion(ht->vertex() != m_diag->topology_traits()->top_right_vertex());
      CGAL_assertion(ht->vertex() != m_diag->topology_traits()->top_left_vertex());
      CGAL_assertion(ht->vertex() != m_diag->topology_traits()->bottom_right_vertex());
      CGAL_assertion(ht->vertex() != m_diag->topology_traits()->bottom_left_vertex());
      CGAL_assertion(VDOL_3::degree(ht->vertex())==3);

      m_top_edges.push_back(ht->next());
      m_top_faces.push_back(ht->next()->opposite()->outer_ccb()->face());

      // m_top_u_values.push_back(ht->vertex()->point().x()); 
      m_top_u_values.push_back(
          ht->next()->opposite()->direction() == ARR_LEFT_TO_RIGHT? 
          ht->next()->opposite()->curve().curve_end_x(ARR_MAX_END):
          ht->next()->opposite()->curve().curve_end_x(ARR_MIN_END));
      
      // vertical curves are only due to intersections with another line 
      if(ht->next()->curve().is_vertical())
        m_intersection_values.push_back(ht->next()->opposite()->curve().x());
 
      ht = ht->opposite()->next()->opposite();
    } 
    m_top_faces.push_back(ht->outer_ccb()->face());
    std::cerr <<" top_vertices.size() " << m_top_u_values.size() <<std::endl;
    std::cerr <<" intersection_values.size() " << m_intersection_values.size() <<std::endl;

  }
  
  template <class ForwardIterator>
  void init_diagram(ForwardIterator begin, ForwardIterator end){
    std::cerr<< " Start Envelope_3 " << std::endl;
    CGAL::Timer timer_1;
    timer_1.start();
    CGAL::lower_envelope_3(begin, end, *m_diag);
    timer_1.stop();
    std::cerr<< " Time: " << timer_1.time() << std::endl;
    print_envelop_diagram(*m_diag);
  }
  
  void init_points(){  
    //std::cerr << " start init points "  << std::endl;
    // we also project on the bounding sphere 
    Poly_int_3 x = shift(Poly_int_3(1),1,0);
    Poly_int_3 y = shift(Poly_int_3(1),1,1);
    Poly_int_3 z = shift(Poly_int_3(1),1,2);
    Poly_int_3 csphere = x*x+y*y+z*z - VOL_CLIPPING_SPHERE; 
    Poly_int_3 tcsphere = VDOL_3::construct_transformed_bisector_3(&svcet_3(),csphere);    

    
    typedef typename Envelope_diagram_2::Vertex_iterator VIT;
 



#if 1   
    // std::cerr << " DRAW VERTECIES "  << std::endl;
    typedef typename Envelope_diagram_2::Surface_const_iterator SIT;
    for(VIT vit = m_diag->vertices_begin(); vit != m_diag->vertices_end();vit++){
      
      Point_3 p_1 = project_back(svcet_3(),vit->point(),vit->surface().transformed_bisector());
      Point_3 p_2 = project_back(svcet_3(),vit->point(),tcsphere);
      Point_3 p_min = (std::min)(p_1,p_2);
      Point_3 p_max = (std::max)(p_1,p_2);
      m_points.push_back(p_min);
      m_points.push_back(p_max);
      for(SIT sit = vit->surfaces_begin(); sit != vit->surfaces_end(); sit++){
        Point_3 p1 = project_back(svcet_3(),vit->point(),(*sit).transformed_bisector());
        m_points.push_back(p1);
        p_min = CGAL::min(p1,p_min);
        p_max = CGAL::max(p1,p_max);
        //  std::cerr << p1 << std::endl;
      }
    
      if(slength(p_max) < VOL_BOUNDING_SPHERE || slength(p_min) < VOL_BOUNDING_SPHERE ){
        for(double t = 0; t < 1; t+= 0.01 ){
          Point_3 p(p_max.x()*(1-t)+p_min.x()*t,p_max.y()*(1-t)+p_min.y()*t,p_max.z()*(1-t) + p_min.z()*t);
          m_points.push_back(p);
        } 
      }
    }
#endif

#if 1 
    // std::cerr << "DRAW SPHERE SURFACE INTERSECTIONS   "  << std::endl;
    typedef typename Envelope_diagram_2::Face_iterator FIT;
    {
      //std::cerr << " COLLECT SURFACES "  << std::endl;
      std::set<Xy_monotone_surface_3> surfaces;
      for(FIT fit = m_diag->faces_begin(); fit != m_diag->faces_end();fit++){
        if(fit->number_of_surfaces() > 0 )
          surfaces.insert(fit->surface());
      }
      //std::cerr << " NUMBER OF BISECTORS "  << std::endl;

      std::vector<typename SVCET_3::X_monotone_curve_2> subcurves;
      std::vector<typename SVCET_3::Curve_2> curves;
      for(typename std::set<Xy_monotone_surface_3>::iterator sit = surfaces.begin(); sit != surfaces.end();sit++){
        Poly_int_2 projection = canonicalize(make_square_free(resultant(sit->transformed_bisector(),tcsphere)));
        curves.push_back(svcet_3().kernel().construct_curve_2_object()(projection));
      }
      typename SVCET_3::Base ck; 
      compute_subcurves (curves.begin(),curves.end(),std::back_inserter(subcurves),false,ck);

      typedef typename std::vector<typename SVCET_3::X_monotone_curve_2>::iterator SCIT;  
      typename Envelope_diagram_2::Face_const_handle   fh;
      for(SCIT scit= subcurves.begin();scit != subcurves.end(); scit++){   
        std::cerr << "."  << std::flush; 
        Point_2 p2 = svcet_3().construct_interior_vertex_2_object()(*scit);
        if(CGAL::assign (fh,m_pl_diag->locate(p2))){
          if(fh->number_of_surfaces() > 0){
            // do not use tcsphere to project back since this is unstable 
            typename Coercion_traits<Poly_int_3,Poly_rat_3>::Cast cast;
            Point_3 B = project_back(svcet_3(),p2,fh->surface().transformed_bisector());
            if(CGAL::abs(CGAL::to_double(canonicalize(cast(csphere)).evaluate(B.z()).evaluate(B.y()).evaluate(B.x()))) < 0.0001){
              project_back(svcet_3(),*scit,fh->surface().transformed_bisector(),std::back_inserter(m_points));
            }
          }
        }
      }
      std::cerr << "."  << std::endl;
    }
    //std::cerr << " DRAW CURVE END "  << std::endl; 

#endif 





#if 1
    // std::cerr << " DRAW EDGES "  << std::endl;
    typedef typename Envelope_diagram_2::Edge_iterator EIT;
    for(EIT eit = m_diag->edges_begin(); eit != m_diag->edges_end();eit++){
      std::cerr << "."  << std::flush;
      project_back(svcet_3(),eit->curve(),eit->surface().transformed_bisector(),std::back_inserter(m_points));
    }
#endif 


#if 0 
    {
      std::vector<Point_3> projected_points; 
      //std::cerr << " DRAW PROJECTED EDGES "  << std::endl;
      typedef typename Envelope_diagram_2::Edge_iterator EIT;
      for(EIT eit = m_diag->edges_begin(); eit != m_diag->edges_end();eit++){
        std::cerr << "."  << std::flush;
        project_back(svcet_3(),eit->curve(),tcsphere,std::back_inserter(projected_points));
      }
      // points are projected on the bounding sphere, but from both sides 
      // remove points that are projected from the outside into interior of the cell 
      for(int i = 0; i < projected_points.size(); i++){
        if(compute_label(projected_points[i]) != 1){
          m_points.push_back(projected_points[i]);
        }
      }
    }
#endif 
    
    //std::cerr << "  DRAW ALL POINTS END "   <<  m_points.size()<< std::endl;  
  }
  
 


  template<class OutputIterator>
  OutputIterator project_back(
      const SVCET_3& svcet_3, 
      const typename SVCET_3::X_monotone_curve_2& xcurve, 
      const Poly_int_3& tbs,
      OutputIterator oit) const {
    
    typename K::Compute_squared_length_3 slength = kernel().compute_squared_length_3_object() ;
    typename K::Construct_vector_3 construct_v = kernel().construct_vector_3_object();
    

    if(xcurve.is_vertical()){
      Coordinate_1 x = xcurve.curve_end_x(ARR_MAX_END);
      Rational y_min,y_max;
      if(xcurve.location(ARR_MIN_END) == ARR_BOTTOM_BOUNDARY){
        y_min = -100000; 
      }else{
        y_min = xcurve.curve_end(ARR_MIN_END).xy().y().upper();
      }
  
      if(xcurve.location(ARR_MAX_END) == ARR_TOP_BOUNDARY){
        y_max =  100000; 
      }else{
        y_max = xcurve.curve_end(ARR_MAX_END).xy().y().lower();
      }     
      
      oit++ = project_back(svcet_3,construct_point_2(&svcet_3,x,y_min),tbs);      
      oit++ = project_back(svcet_3,construct_point_2(&svcet_3,x,y_max),tbs);

      if( CGAL::sign(y_min) * CGAL::sign(y_max) == CGAL::NEGATIVE ){
        oit = split_vertical_arc(svcet_3,x,y_min,Rational(0), tbs,VOL_POINT_GENERATION,oit);
        oit = split_vertical_arc(svcet_3,x,Rational(0),y_max,tbs,VOL_POINT_GENERATION, oit);
      }else{      
        oit = split_vertical_arc(svcet_3,x,y_min,y_max,tbs,VOL_POINT_GENERATION, oit);
      }
      return oit;
    }

    
    Coordinate_1 x_min,x_max;
    Point_3 p_min_3, p_max_3; 
    
    Arr_parameter_space min_loc = xcurve.location(ARR_MIN_END);
    if(min_loc == ARR_INTERIOR ){
      x_min =  xcurve.curve_end_x(ARR_MIN_END);
      //p_min_3 = project_back(svcet_3,xcurve.curve_end(ARR_MIN_END),tbs);
    }else{
      if( min_loc == ARR_BOTTOM_BOUNDARY || min_loc == ARR_TOP_BOUNDARY){
        x_min = xcurve.curve_end_x(ARR_MIN_END);
        x_min = Coordinate_1(ckernel().kernel().approximate_relative_1_object()(x_min,53).second+0.000001);
      }else{
        x_min = Coordinate_1(-110);
      }
      //p_min_3 = project_back(svcet_3,Point_2(x_min,xcurve.curve(),xcurve.arcno()),tbs);
    }

    Arr_parameter_space max_loc = xcurve.location(ARR_MAX_END);
    if(max_loc == ARR_INTERIOR ){
      x_max =  xcurve.curve_end_x(ARR_MAX_END);
      //p_max_3 = project_back(svcet_3,xcurve.curve_end(ARR_MAX_END),tbs);
    }else{
      if( max_loc == ARR_BOTTOM_BOUNDARY || max_loc == ARR_TOP_BOUNDARY){
        x_max = xcurve.curve_end_x(ARR_MAX_END);
        x_max = Coordinate_1(ckernel().kernel().approximate_relative_1_object()(x_max,53).first-0.000001);
      }else{  
        x_max = Coordinate_1(110);
      }
      //p_max_3 = project_back(svcet_3,Point_2(x_max,xcurve.curve(),xcurve.arcno()),tbs); 
    }
    if(x_min > x_max) {
//      std::cerr << "I had to skip an edge because it seems too small" << std::endl;      
//      std::cerr << to_double(x_min) << " " << to_double(x_max) << std::endl;
      return oit; // in case an edge is too small 
    }
    
    
    if(!xcurve.is_in_x_range_interior(x_min)){
      p_min_3 = project_back(svcet_3,xcurve.curve_end(ARR_MIN_END),tbs);
    }else{
      p_min_3 = project_back(svcet_3,Point_2(x_min,xcurve.curve(),xcurve.arcno()),tbs); 
    }
    
    if(!xcurve.is_in_x_range_interior(x_max)){
      p_max_3 = project_back(svcet_3,xcurve.curve_end(ARR_MAX_END),tbs);
    }else{
      p_max_3 = project_back(svcet_3,Point_2(x_max,xcurve.curve(),xcurve.arcno()),tbs); 
    }
 
    *oit++ = p_min_3;
    *oit++ = p_max_3; 

  
    if( CGAL::sign(x_min) * CGAL::sign(x_max) == CGAL::NEGATIVE ){      
      Coordinate_1 x_mid(0);
      Point_3 p_mid_3(project_back(svcet_3,Point_2(x_mid,xcurve.curve(),xcurve.arcno()),tbs)); 
      oit =  split_arc(svcet_3,xcurve,tbs,x_min,x_mid,p_min_3,p_mid_3,VOL_POINT_GENERATION,oit);
      oit =  split_arc(svcet_3,xcurve,tbs,x_mid,x_max,p_mid_3,p_max_3,VOL_POINT_GENERATION,oit);
    }else{
      oit =  split_arc(svcet_3,xcurve,tbs,x_min,x_max,p_min_3,p_max_3,VOL_POINT_GENERATION,oit);
    }
    return oit; 
  }
  

 template <class OutputIterator>
  OutputIterator split_vertical_arc(
      const SVCET_3& svcet_3, 
      const Coordinate_1 x,
      const Rational& y_min,
      const Rational& y_max, 
      const Poly_int_3& tbs,
      const double& sdistance, 
      OutputIterator oit) const  
  {
    Rational y_mid((y_min+y_max)/2);

    Point_3 p_min_3 = project_back(svcet_3,construct_point_2(&svcet_3,x,y_min),tbs);
    Point_3 p_mid_3 = project_back(svcet_3,construct_point_2(&svcet_3,x,y_mid),tbs);
    Point_3 p_max_3 = project_back(svcet_3,construct_point_2(&svcet_3,x,y_max),tbs);
    
    if( kernel().compute_squared_distance_3_object()(p_min_3,p_max_3) < sdistance)
      return oit++ =  p_min_3;
    if( to_double(y_max)-to_double(y_min) < 0.00001 ) 
      return oit++ =  p_min_3;

    bool is_visible_min = (slength(p_min_3) <= VOL_BOUNDING_SPHERE );
    bool is_visible_mid = (slength(p_mid_3) <= VOL_BOUNDING_SPHERE );
    bool is_visible_max = (slength(p_max_3) <= VOL_BOUNDING_SPHERE );
    
    if(is_visible_min || is_visible_mid) 
        split_vertical_arc(svcet_3,x,y_min,y_mid,tbs,VOL_POINT_GENERATION, oit);
    if(is_visible_mid || is_visible_max)
        split_vertical_arc(svcet_3,x,y_mid,y_max,tbs,VOL_POINT_GENERATION, oit);
    return oit;
  }




  template <class OutputIterator>
  OutputIterator split_arc(
      const SVCET_3& svcet_3, 
      const typename SVCET_3::X_monotone_curve_2& xcurve, 
      const Poly_int_3& tbs,
      const Coordinate_1& x_min,
      const Coordinate_1& x_max, 
      const Point_3& p_min_3, 
      const Point_3& p_max_3,
      const double& sdistance, 
      OutputIterator oit) const {
    
    
    Coordinate_1 x_mid(ckernel().kernel().bound_between_1_object()(x_min,x_max));
    Point_2 p_mid_2 (x_mid,xcurve.curve(),xcurve.arcno());
    Point_3 p_mid_3 = project_back(svcet_3,p_mid_2,tbs);
  
    if(kernel().compute_squared_distance_3_object()(p_min_3,p_max_3) < sdistance  || to_double(x_max)-to_double(x_min) < 0.0000001) {
      return *oit++ = p_min_3;
    }
    
    bool is_visible_min = (slength(p_min_3) <= VOL_BOUNDING_SPHERE );
    bool is_visible_mid = (slength(p_mid_3) <= VOL_BOUNDING_SPHERE );
    bool is_visible_max = (slength(p_max_3) <= VOL_BOUNDING_SPHERE );
    
    if(is_visible_min || is_visible_mid)
      oit = split_arc(svcet_3,xcurve,tbs,x_min,x_mid,p_min_3,p_mid_3,sdistance,oit);
    if(is_visible_mid || is_visible_max)
      oit = split_arc(svcet_3,xcurve,tbs,x_mid,x_max,p_mid_3,p_max_3,sdistance,oit);
    return oit;
    
  }
  
  Point_3 project_back(
      const SVCET_3& svcet_3, 
      const typename SVCET_3::Point_2& point_2, 
      const Poly_int_3& tbs) const {
    
    std::pair<double,double> approx = point_2.xy().to_double();
    Rational u(approx.first);
    Rational v(approx.second);
    Rational r; 

    typename CGAL::Coercion_traits<Poly_int_3,Poly_rat_3>::Cast cast; 
    Poly_rat_3 poly_3 = cast(CGAL::swap(tbs,0,2));
    Poly_rat_1 poly_1 = CGAL::canonicalize(CGAL::evaluate(CGAL::evaluate(poly_3,Poly_rat_2(u)),Poly_rat_1(v)));
    CGAL_postcondition(CGAL::is_zero(poly_1) || poly_1.lcoeff()==1);

    switch(CGAL::degree(poly_1)){
    case 0: 
      // return Point_3(1000,1000,1000);
      r = 0; 
      break; 
    case 1: 
      r = -poly_1[0]/poly_1[1];
      // bisector is just a plane and the point is on the wrong side
      // or the point is on the infinite line at infinity. 
      if(r < 0) r = 0; // return Point_3(1000,1000,1000);
      break; 
    default: // = 2 
      CGAL_postcondition(CGAL::degree(poly_1) == 2); 
      Rational p = poly_1[1];
      Rational q = poly_1[0];
      Rational discr = p*p/4 - q;
      if(discr < 0) r = 10; //  Point_3(1000,1000,1000);
#if 0
      typedef typename Get_arithmetic_kernel<Rational>::Arithmetic_kernel AT;
      typedef typename AT::Field_with_sqrt FWS;
      FWS sdiscr = CGAL::sqrt(FWS(CGAL::abs(discr)));
      r = to_double(-p/2 + sdiscr);
#else     
      double sdiscr = CGAL::sqrt(CGAL::to_double(CGAL::abs(discr)));
      r = Rational(-p/2 + sdiscr);
#endif 
     
    }
     

    EVector_3 v1 = ekernel().construct_scaled_vector_3_object()(svcet_3.v1(),u);
    EVector_3 v2 = ekernel().construct_scaled_vector_3_object()(svcet_3.v2(),v*r);
    EVector_3 v3 = ekernel().construct_scaled_vector_3_object()(svcet_3.v3(),r);
    EPoint_3 ep = svcet_3.origin();
    ep = ekernel().construct_translated_point_3_object()(ep,v1);
    ep = ekernel().construct_translated_point_3_object()(ep,v2);
    ep = ekernel().construct_translated_point_3_object()(ep,v3);
      
    Point_3 point_3(CGAL::to_double(ep.x()),CGAL::to_double(ep.y()),CGAL::to_double(ep.z()));
    
    return point_3; 
  }
  

  return_type compute_label(Point_3 p) const {
    return compute_label(EPoint_3(p.x(),p.y(),p.z()));
  }
  
  EPoint_3 in_local_coordinates(const Point_3& p) const {
    return in_local_coordinates(EPoint_3(p.x(),p.y(),p.z()));
  }

  EPoint_3 in_local_coordinates(const EPoint_3& ep) const {
    EVector_3 ev_diff = ekernel().construct_vector_3_object()(svcet_3().origin(),ep);
    Rational  lv1 = ekernel().compute_squared_length_3_object()(svcet_3().v1());
    Rational  lv2 = ekernel().compute_squared_length_3_object()(svcet_3().v2());
    Rational  lv3 = ekernel().compute_squared_length_3_object()(svcet_3().v3());
    
    Rational  c1 = ekernel().compute_scalar_product_3_object()(svcet_3().v1(),ev_diff) / lv1; 
    Rational  c2 = ekernel().compute_scalar_product_3_object()(svcet_3().v2(),ev_diff) / lv2; 
    Rational  c3 = ekernel().compute_scalar_product_3_object()(svcet_3().v3(),ev_diff) / lv3; 
    return EPoint_3(c1,c2,c3);
  }

  typename SVCET_3::Point_2 project_point(const Point_3& p) const {
    return project_point(EPoint_3(p.x(),p.y(),p.z()));
  }
  typename SVCET_3::Point_2 project_point(const EPoint_3& ep) const {
    EPoint_3 lep = in_local_coordinates(ep);
    CGAL_precondition(lep.z() != 0);
    Rational u = lep.x(); 
    Rational v = lep.y()/lep.z();
    return CGAL::construct_point_2(&svcet_3() ,Rational(u),Rational(v));
  }

  
  return_type compute_label(EPoint_3 ep, bool b = true ) const {
    
    
    typename SVCET_3::Point_2 p_2 = project_point(ep);
    
    boost::optional<const Xy_monotone_surface_3&> xy_surface_3 = 
      VDOL_3::compute_surface(*m_pl_diag,p_2);
    
    if (! xy_surface_3.is_initialized() ){
      return 1;
    }
    
    
    if(
        CGAL::sign_at(  xy_surface_3.get().bisector(),
            ekernel().construct_cartesian_const_iterator_3_object()(ep),
            ekernel().construct_cartesian_const_iterator_3_object()(ep,0))
        ==
        CGAL::sign_at(  xy_surface_3.get().bisector(),
            ekernel().construct_cartesian_const_iterator_3_object()(svcet_3().origin()),
            ekernel().construct_cartesian_const_iterator_3_object()(svcet_3().origin(),0)))
      {
        //std::cerr << " Label 1 "  << std::endl; 
        return 1;
      }
    else
      {
        //std::cerr << " Label "  << xy_surface_3.get().id()<< std::endl; 
        //return 0; 
        return xy_surface_3.get().id();
      }
    
    return 0; 
  }

  
  // Sphere_3 bounding_sphere(){return Sphere_3(Point_3(0,0,0),VOL_BOUNDING_SPHERE);}
  double slength(const Point_3& p) const {return p.x()*p.x()+ p.y()*p.y()+p.z()*p.z(); }
 
}; 
  

//  MEMBERS 
  Half_voronoi_cell_of_line_3 m_hvcl_3_pos;
  Half_voronoi_cell_of_line_3 m_hvcl_3_neg;
  std::vector<std::pair<Point_3,double> > m_wpoints; 
public: 

  const Envelope_diagram_2& envelope_diagram_2(bool pos) const {
    return (pos)?m_hvcl_3_pos.diag():m_hvcl_3_neg.diag();
  }
  long number_of_vertical_asymptotes() const {
    return
      m_hvcl_3_pos.number_of_vertical_asymptotes()+  
      m_hvcl_3_neg.number_of_vertical_asymptotes(); 
  }


  template <class ForwardIterator>
  Voronoi_cell_of_line_3(int seed, ForwardIterator begin, ForwardIterator end)
    :m_hvcl_3_pos(SVCET_3(*begin++,seed),begin,end),
     m_hvcl_3_neg(m_hvcl_3_pos.svcet_3().mirror(),begin,end)
  {
    // std::cerr <<"WITH SEED: "<< seed <<std::endl;
    
    std::vector<Point_3> points_raw; 
    std::copy(m_hvcl_3_pos.points().begin(),m_hvcl_3_pos.points().end(), std::back_inserter(points_raw));
    std::copy(m_hvcl_3_neg.points().begin(),m_hvcl_3_neg.points().end(), std::back_inserter(points_raw));
    
    //std::cerr << " points_raw "  << points_raw.size() << std::endl; 
   
    // filter points such that the distance is ok
    // vertieces in the arrangement remain since they are the first points in points_raw 
    typename K::Compute_squared_distance_3 sdist = m_hvcl_3_pos.kernel().compute_squared_distance_3_object();    

    std::vector<Point_3> points;
    for(int i = 0; i < points_raw.size(); i++){
      if( slength(points_raw[i]) < VOL_CLIPPING_SPHERE + 0.1){
        bool insert = true; 
        for(int j = 0; j < points.size() && insert ; j++){
          if(CGAL::to_double(sdist(points_raw[i],points[j])) < VOL_POINT_DIST){
            insert = false ;
          }          
        }
        if(insert) points.push_back(points_raw[i]);
      }
    }

    //std::cerr << " points "  << points.size() << std::endl; 

    double sdistance = 1000; 
    for(int i = 0; i < points.size(); i++){
      for(int j = i+1; j<points.size();j++){
        sdistance = std::min(CGAL::to_double(sdist(points[i],points[j])),sdistance);
      }
    }
    //std::cerr << " sdistance "  << sdistance << std::endl; 


    for(typename std::vector<Point_3>::iterator it = points.begin(); it != points.end(); it++){
      if( slength(*it) <= VOL_CLIPPING_SPHERE + 0.1){
        m_wpoints.push_back(std::pair<Point_3,double>(*it,VOL_POINT_WEIGHT));
      }
    }
    //std::cerr << " wpoints "  << m_wpoints.size() << std::endl; 
  }

  double slength(const Point_3& p) const {return p.x()*p.x()+ p.y()*p.y()+p.z()*p.z(); }


  Sphere_3 bounding_sphere(){return Sphere_3(Point_3(0,0,0),VOL_BOUNDING_SPHERE);}
  
  return_type operator()(typename K::Point_3 p, bool b = true ) const {
    return compute_label(p);
  }


  // use positive side 
  bool compute_side(const EPoint_3& ep) const {
    EK ek =  m_hvcl_3_pos.ekernel();
    SVCET_3 svcet_3 = m_hvcl_3_pos.svcet_3();
    EVector_3 ev_diff =ek.construct_vector_3_object()(svcet_3.origin(),ep);
    Sign      s1  = CGAL::sign(ek.compute_scalar_product_3_object()(svcet_3.v3(),ev_diff));
    Sign      s2  = CGAL::sign(ek.compute_scalar_product_3_object()(svcet_3.v2(),ev_diff));
    if(s1 == CGAL::POSITIVE) return true;
    if(s1 == CGAL::NEGATIVE) return false;
    if(s2 == CGAL::POSITIVE) return true;
    if(s2 == CGAL::NEGATIVE) return false;
    return true;
  }

 
  template<class OutputIterator>
  bool contains(const EPoint_3& ep, OutputIterator oit) const {
#ifdef CGAL_VDOL_USE_COUNTER
    count_cell_contains++;
#endif

    return compute_side(ep)? 
      m_hvcl_3_pos.contains(ep , oit):
      m_hvcl_3_neg.contains(ep , oit);
  }
  

//   int compute_label(typename K::Point_3 p) const {
//     if( slength(p) > VOL_CLIPPING_SPHERE  ) return 0;
//     EPoint_3 ep(p.x(),p.y(),p.z());
//     CGAL::Sign sign = compute_side(ep);
    
//     if (sign == CGAL::ZERO){  
//       if(CGAL::is_zero(CGAL::squared_distance(ep,m_hvcl_3_pos.svcet_3().line())))
//         return 1; 
//       //std::cerr << " PERTURB POINT QUERY ON DISCONTINUITY PLANE ! " << std::endl; 
//       //std::cerr << p << std::endl;
//       ep = EPoint_3(p.x()+0.000021,p.y()-0.00002,p.z()+0.00000001);
//     }
    
//     //m_hvcl_3_pos.compute_segment_index(p);
//     return (sign == POSITIVE) ? m_hvcl_3_pos.compute_label(ep): m_hvcl_3_neg.compute_label(ep);
//   }
  
  int compute_segment_index(const typename K::Point_3& p) const {
    return m_hvcl_3_pos.compute_segment_index(p);
  }


  const std::vector<std::pair<Point_3,double> >& wpoints() const {return m_wpoints;}
  std::vector<std::pair<Point_3,double> >& wpoints(){return m_wpoints;}
  const Line_3& line_3() const {return m_hvcl_3_pos.svcet_3().base_line_3();}

};
}
 
#endif // CGAL_VORONOI_CELL_OF_LINE_3_H
