#define CGAL_AK_ENABLE_DEPRECATED_INTERFACE 1 

// ----------------------------------------------------
// includes for VOL_3
#include <CGAL/basic.h>
#include <CGAL/Timer.h>
#include <CGAL/Cartesian.h>
#include <CGAL/macros.h>

#include <CGAL/Arithmetic_kernel.h>
#include <CGAL/Algebraic_kernel_1.h>
#include <CGAL/Algebraic_curve_kernel_2_generator.h> // generates a default kernel 
#include <CGAL/Curved_kernel_via_analysis_2/Curved_kernel_via_analysis_2_impl.h> // traits for Arr_2

#include <CGAL/VDOL_3/Single_voronoi_cell_envelope_traits_3.h>

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
//#include <CGAL/Exact_predicates_exact_constructions_kernel.h>

// -----------------------------------------------------
// includes from /demo/VDOL_3/include 
#include <CGAL/Voronoi_diagram_of_lines_3.h> 


// get arithmetic kernel 
typedef CGAL::Arithmetic_kernel AK;
typedef AK::Integer Integer;
typedef AK::Rational Rational;

// define linear kernel 
#if 1
typedef CGAL::Cartesian<Rational> EK_3;
struct K: public CGAL::Exact_predicates_inexact_constructions_kernel {};
#else
typedef CGAL::Cartesian<Rational> EK_3;
typedef EK_3 K;
// struct K: public CGAL::Exact_predicates_exact_constructions_kernel {};
#endif

typedef EK_3::Line_3                                    ELine_3;
typedef EK_3::Point_3                                   EPoint_3;

// define algebraic kerenel 
typedef CGAL::Algebraic_kernel_1_generator<Integer>     GEN_AK_1;
typedef GEN_AK_1::Default_algebraic_kernel_1            AK_1;
typedef CGAL::Algebraic_curve_kernel_2<AK_1>            AK_2;

// define curved kernel 
typedef CGAL::Curved_kernel_via_analysis_2< AK_2 >      CK_2;

// define traits for lower envelop diagram 
typedef CGAL::VDOL_3::
Single_voronoi_cell_envelope_traits_3<EK_3, CK_2>       SVCET_3;
 

// Domain: DB
typedef CGAL::Voronoi_diagram_of_lines_3<SVCET_3,K>   VDOL_3;



void copy_file_into(const char *infilename, const char *outfilename){
  std::ifstream     in_file (infilename);
  std::ofstream     out_file (outfilename);
  std::cout << "save " << infilename << " to " << outfilename << "... " << std::flush;
  if (! in_file.is_open()) {
    std::cerr << "Failed to open " << infilename << "!" << std::endl;
    return; 
  }
  if (! out_file.is_open()) {
    std::cerr << "Failed to open " << outfilename << "!" << std::endl;
    return; 
  }
  out_file << in_file.rdbuf();
  std::cout << "  done "<< std::endl;
}

template <typename OutputIterator>
OutputIterator read_lines(const char *filename, OutputIterator oi){

  // Open the input file.
  std::ifstream     in_file (filename);
  if (! in_file.is_open()) {
    std::cerr << "Failed to open " << filename << "!" << std::endl;
    return oi;
  }
  
  // Read the lines from the file
  // The input file format should be (all coordinate values are integers):
  // <n>                                       // number of lines.
  // <a1_x> <a1_y> <a1_z> <b1_x> <b1_y> <b1_z> // line #1.
  // <a2_x> <a2_y> <a2_z> <b2_x> <b2_y> <b2_z> // line #2.
  //   :      :       :      :
  // <an_x> <an_y> <an_z> <bn_x> <bn_y> <bn_z> // line #n.
  
  // read number of lines 
  unsigned int n; in_file >> n;

  for (int k = 0; k < n; ++k) {
    int a,b,c,d,e,f;
    in_file >> a >> b >> c>> d >> e >> f;
    typedef Rational FT; 
    *oi++ = ELine_3(EPoint_3(FT(a),FT(b),FT(c)),EPoint_3(FT(d),FT(e),FT(f)));
  }
  in_file.close();
}

template<typename Line, typename Point> inline 
CGAL::Comparison_result compare_squared_distance(const Line& l1, const Line& l2, const Point& p){
  return CGAL::compare(
      CGAL::squared_distance(l1,p),
      CGAL::squared_distance(l2,p));
}

int main(int argc, char **argv)
{

  std::cout << "THIS IS A NEW COMPILATION " << std::endl;

// ---------------------------------------- 
// the voronoi cell of one line 
  
  //CGAL::set_pretty_mode(std::cout);
  //CGAL::set_pretty_mode(std::cerr);
  
  // Get the name of the input file from the command line, or use the default
  // last.dat file if no command-line parameters are given.
  const char * filename = (argc >= 2) ? argv[1] : "last.dat"; 
  if(argc > 1){
    copy_file_into(filename,"last.dat");
  }

  std::cout << "Run test with file: " << filename << std::endl;

  std::vector<SVCET_3::Surface_3> lines;
  read_lines(filename,std::back_inserter(lines));
  if(lines.size()==0) 
    std::cerr << "Failed to read file " << filename << std::endl;
  
  std::cout << "Number of lines (incl base line): "<< lines.size() << std::endl;
  

  VDOL_3 vdol_3(lines.begin(),lines.end());
  
  assert(lines.size()>=2);
  
  for(int i = -10; i<10; i+=1){ 
    for(int j = -10; j<10; j+=1){ 
      for(int k = -10; k<10; k+=1){
        EPoint_3 p(i,j,k);
        int min_l = compare_squared_distance(lines[0],lines[1],p)==CGAL::SMALLER?0:1;
        for(int l = 0; l < lines.size(); l++){
          if(compare_squared_distance(lines[l],lines[min_l],p) == CGAL::SMALLER){
            min_l=l;
          }
        }
        std::vector<int> result_1; 
        for(int l = 0; l < lines.size(); l++){
          if(compare_squared_distance(lines[min_l],lines[l],p)==CGAL::EQUAL){
            result_1.push_back(l);
          }
        }  
        std::vector<int> result_2;
        vdol_3.locate(p,std::back_inserter(result_2));
           
//         for(int l =0; l < result_1.size();l++){
//           std::cout << result_1[l] << " ";
//         }
//         std::cout<<std::endl;
//         for(int l =0; l < result_2.size();l++){
//           std::cout << result_2[l] << " ";
//         }
//         std::cout<<std::endl;
//         std::cout<<std::endl;
        assert(result_1 == result_2);
      }
    }
  }


  return 0;
}
